# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3
identity = lambda x: x
default_encoding = 'UTF-8'

if PY3:
    # reduce no longer a built-in
    from functools import reduce
    import io as StringIO
    import queue
    import decimal

    basestring = (str, bytes)
    xrange = range
    iterbytes = identity

    def ensure_bytes(s):
        if type(s) == str:
            return bytes(s, default_encoding)
        elif type(s) == float:
            return ensure_bytes(str(s))
        elif type(s) == int:
            return ensure_bytes(str(s))
        else:
            return bytes(s)

    def ensure_str(s):
        if type(s) == bytes:
            return s.decode(default_encoding)
        elif type(s) == str:
            return s
        else:
            return s

    def escape_string(s):
        if isinstance(s, str):
            s = s.encode('unicode_escape').decode(default_encoding)
        elif isinstance(s, bytes):
            s = s.encode('unicode_escape')
        return s

    def unescape_string(s):
        if isinstance(s, str):
            s = s.encode(default_encoding).decode('unicode_escape')
        elif isinstance(s, bytes):
            s = s.decode('unicode_escape')
        return s
else:
    try:
        import cdecimal as decimal
    except ImportError:
        import decimal
    try:
        import cStringIO as StringIO
    except ImportError:
        import StringIO
    import Queue as queue

    # Bytes object in Py3 does not require a call to ord to return a
    # numeric type 
    iterbytes = lambda x: (ord(c) for c in x)

    input = raw_input

    ensure_bytes = lambda s: bytes(s)
    ensure_str = identity

    def escape_string(s):
        if isinstance(s, basestring):
            s = s.encode('string_escape')
        return s

    def unescape_string(s):
        if isinstance(s, basestring):
            s = s.decode('string_escape')
        return s



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from .constants import *
from .errors import *

from . import _teradata
from . import _teradatapt
from .config import Config, message_write_default
from .encrypt import create_key_file
from .io import home_file
from .logging import colors, log
from .parser import Argument, Command
from .utils import prompt_bool


class MainCommand(Command):
    name = "giraffez"
    description = ascii_giraffe
    usage = "giraffez [args] <command> [options]"

    global_arguments = [
        Argument("-D", "--dsn"),
        Argument("-c", "--conf"),
        Argument("-k", "--key"),
        Argument("-v", "--verbose", action="count", default=0, dest="log_level"),
        Argument("-h", "--help", default=False, help="Print usage"),
    ]

    arguments = [
        Argument("-V", "--version", action="version", help="Print version"),
        Argument("--protect", default=False, help=("Use connection locking to protect Teradata "
            "accounts from locking from bad attempts")),
    ]

    commands = [
        "commandline.CmdCommand",
        "commandline.ConfigCommand",
        "commandline.ExportCommand",
        "commandline.FmtCommand",
        "commandline.InsertCommand",
        "commandline.LoadCommand",
        "commandline.RunCommand",
        "commandline.ShellCommand",
        "commandline.SecretCommand",
    ]

    default = False

    def run(self, test_args=None):
        args = self.parse_args(test_args)
        log.level = args.log_level
        log.debug(sys.version.replace("\n", ""))
        if args.module_name is None:
            self.print_help()
            self.exit()
        if args.module_name == "secret" or (args.module_name == "config" and args.init):
            args.run(args)
            return
        try:
            with Config(conf=args.conf, key_file=args.key) as config:
                colorful = config.get_value("colorful", default=True)
                if not colorful:
                    colors.is_colorful = False
            args.run(args)
        except ConfigNotFound as error:
            log.debug(error)
            if args.conf is None:
                args.conf = home_file(".girafferc")
            if prompt_bool(("Configuration file '{}' not found, would you like to create one now "
                    "using the default template?").format(args.conf), default=True):
                result = Config.write_default(args.conf)
                if result:
                    log.write(colors.green(result))
                    log.write(message_write_default.format(args.conf))
                    self.run()
                else:
                    log.write(colors.fail("Was not successful"))
        except KeyNotFound as error:
            log.debug(error)
            if args.key is None:
                args.key = home_file(".giraffepg")
            if prompt_bool("Key file '{}' not found, would you like to create one now?".format(args.key),
                    default=True):
                create_key_file(args.key)
                log.write("Key file '{}' created successfully.".format(args.key))
                self.run()
            else:
                raise error
        except (_teradata.InvalidCredentialsError, _teradatapt.InvalidCredentialsError) as error:
            if args.protect:
                Config.lock_connection(args.conf, args.dsn, args.key)
            raise error



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

"""
giraffez
~~~~~~~~~~~~

A Teradata client making use of CLIv2 and TPT API. It is made to be
user-friendly and very fast.

:copyright: (c) 2016 by Capital One Services, LLC. See AUTHORS for more details.
:license: Apache 2.0, see LICENSE for more details.

"""

__title__ = 'giraffez'
__version__ = '2.0.8'
__authors__ = ['Christopher Marshall', 'Kyle Travis']
__license__ = 'Apache 2.0'
__all__     = ['BulkExport', 'BulkLoad', 'Cmd', 'Config', 'Secret']


try:
    from . import _teradata
    from . import _teradatapt
except ImportError as error:
    raise Exception("""{}.

This indicates that either the giraffez C extensions did not compile
correctly, or more likely, there is an issue with the environment or
installation of the Teradata dependencies. Both the Teradata Call-Level
Interface Version 2 and Teradata Parallel Transporter API require
several environment variables to be set to find the shared library
files and error message catalog.

For more information, refer to this section in the giraffez
documentation:
    http://www.capitalone.io/giraffez/intro.html#environment.
""".format(error))

from ._teradata import TeradataError
from ._teradatapt import (
    EncoderError,
    TeradataError as TeradataPTError
)
from .cmd import Cmd
from .config import Config
from .constants import SILENCE, VERBOSE, DEBUG, INFO
from .errors import (
    GiraffeError,
    GiraffeTypeError,
    InvalidCredentialsError,
    ConnectionLock
)
from .encoders import TeradataEncoder as Encoder
from .export import BulkExport
from .io import (
    Reader,
    Writer
)
from .load import BulkLoad
from .logging import log, setup_logging
from .secret import Secret
from .types import Column, Columns, Date, Decimal, Time, Timestamp
from .utils import register_graceful_shutdown_signal


"""
Module that monkey-patches json module when it's imported so
JSONEncoder.default() automatically checks for a special "to_json()"
method and uses it to encode the object if found. This is done in 
__init__ to ensure that the patch is applied before the JSONEncoder
is used. This is mainly to work with the custom date types when
being serialized into json.
"""
from json import JSONEncoder

def _default(self, obj):
    return getattr(obj.__class__, "__json__", _default.default)(obj)

_default.default = JSONEncoder().default  # Save unmodified default.
JSONEncoder.default = _default # replacement

# Set default logging handler to avoid "No handler found" warnings.
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import json
import time
import argparse
import itertools
import subprocess
import shlex
import re
import io

import yaml

try:
    from . import _teradata
except ImportError:
    pass

from .constants import *
from .errors import *

from .config import Config, message_write_default
from .cmd import TeradataCmd
from .encoders import null_handler, TeradataEncoder
from .encrypt import create_key_file
from .export import TeradataBulkExport
from .fmt import format_table, replace_cr
from .io import ArchiveFileReader, FileReader, Reader, Writer, \
    file_delimiter, file_exists, home_file
from .logging import colors, log, setup_logging
from .load import TeradataBulkLoad
from .parser import Argument, Command
from .shell import GiraffeShell
from .sql import parse_statement
from .utils import pipeline, prompt_bool, readable_time, \
    register_shutdown_signal, register_graceful_shutdown_signal, \
    show_warning, timer

from ._compat import *


setup_logging()

# Registers the standard shutdown handler using C signals. Since parts
# of the giraffez code will execute without acquiring the GIL, Python's
# normal Ctrl+C behavior will not work. This signal handler attempts
# to remedy this by setting a signal handler in C that imitates the
# desired behavior. The goal is to avoid having to acquire the GIL but
# this could be changed in the future if the cost of acquiring the GIL
# in the C extension is shown to be minimal.
register_shutdown_signal()


class CmdCommand(Command):
    name = "cmd"
    description = "Execute Teradata commands"
    usage = "giraffez cmd <query> [options]"
    help = "Execute commands in Teradata (using TDCLIV2)"

    arguments = [
        Argument("query", help="Query command (string or file)"),
        Argument("-t", "--table-output", default=False, help="Print in table format"),
        Argument("-j", "--json", default=False, help="Export in json format"),
        Argument("-d", "--delimiter", default='\t', help="Text delimiter"),
        Argument("-n", "--null", default='NULL', help="Set null character"),
        Argument("--header", default=False, help="Do prepend header"),
    ]

    def run(self, args):
        if args.json:
            args.header = False
        elif args.table_output:
            args.header = True
        args.delimiter = unescape_string(args.delimiter)
        start_time = time.time()
        register_graceful_shutdown_signal()
        with TeradataCmd(log_level=args.log_level, config=args.conf, key_file=args.key,
                dsn=args.dsn) as cmd:
            cmd.options("table_output", args.table_output)
            with Writer() as out:
                result = cmd.execute(args.query, header=args.header)
                if out.is_stdout:
                    log.info(colors.green(colors.bold("-"*32)))
                if args.json:
                    for row in result.to_dict():
                        out.writen(json.dumps(row))
                elif args.table_output and not (args.query.lower().startswith("show") or \
                        args.query.lower().startswith("help")):
                    rows = [replace_cr(row) for row in result]
                    out.writen("\r{}".format(format_table(rows)))
                else:
                    for row in result:
                        out.writen(args.delimiter.join([replace_cr(str(x)) for x in row]))
                if out.is_stdout:
                    log.info(colors.green(colors.bold("-"*32)))
                count = sum([n.count for n in result.statements])
                log.info("Results", "{} rows in {}".format(count, readable_time(time.time() - start_time)))


class ConfigCommand(Command):
    name = "config"
    description = "giraffez configuration"
    usage = "giraffez config <options>"
    help = "View/Edit giraffez configuration"

    arguments = [
        Argument("-n", "--no-newline", action="store_true",
            help="Do not output a newline with value from --get"),
        Argument("-d", "--decrypt", action="store_true",
            help="Decrypt the contents of the settings when using --list"),

        Argument("--init", action="store_true",
            help="Initialize a new configuration file at [--config]", group="mutual"),
        Argument("--get", metavar="key",
            help="Retrieve the key at the given path, separating nested keys with '.'",
            group="mutual"),
        Argument("-l", "--list", action="store_true",
            help="Show the contents of the configuration file", group="mutual"),
        Argument("--set", nargs=2, metavar=("key", "value"),
            help="Set the key to the given value", group="mutual"),
        Argument("--unset", metavar="key", help="Unset the given key", group="mutual"),
        Argument("--unlock", metavar="dsn",
            help="Unlock the connection with the given DSN", group="mutual"),
    ]

    def run(self, args):
        if args.conf is None:
            args.conf = home_file(".girafferc")
        if args.key is None:
            args.key = home_file(".giraffepg")
        if args.init:
            if file_exists(args.conf):
                if not prompt_bool(("Configuration file '{}' found, would you like to overwrite "
                    "it with defaults?").format(args.conf), default=False):
                    return
            result = Config.write_default(args.conf)
            if result:
                log.write(colors.green(result))
                log.write(message_write_default.format(args.conf))
            else:
                log.write(colors.fail("Was not successful"))
            create_key_file(args.key)
            log.write("Key file '{}' created successfully.".format(args.key))
        elif args.get is not None:
            key = args.get
            with Config(conf=args.conf, key_file=args.key) as c:
                value = c.get_value(key)
            if value == -1:
                log.write("{}: not set".format(key))
            else:
                if not isinstance(value, str):
                    value = json.dumps(value, indent=4, sort_keys=True)
                if not args.no_newline:
                    value = "{}:\n{}\n".format(key, value)
                log.write(value)
        elif args.list:
            with Config(conf=args.conf, key_file=args.key) as c:
                log.write(c.list_value(args.decrypt))
        elif args.set is not None:
            key, value = args.set
            with Config(conf=args.conf, mode="w", key_file=args.key) as c:
                c.set_value(key, value)
                c.write()
        elif args.unset is not None:
            with Config(conf=args.conf, mode="w", key_file=args.key) as c:
                c.unset_value(args.unset)
                c.write()
        elif args.unlock is not None:
            Config.unlock_connection(args.conf, args.unlock, args.key)


class ExportCommand(Command):
    name = "export"
    description = "Export data from Teradata"
    usage = "giraffez export <query> [output_file] [options]"
    help = "Export data from Teradata"

    arguments = [
        Argument("query", help="Table or query to export"),
        Argument("output_file", nargs="?", help="Output file or pipe"),
        Argument("-a", "--archive", default=False, help="Export in archive format (packed binary)"),
        Argument("-z", "--gzip", default=False, help="Use gzip compression (only for archive)"),
        Argument("-j", "--json", default=False, help="Export in json format"),
        Argument("-d", "--delimiter", default="|", help="Text delimiter"),
        Argument("-n", "--null", default="NULL", help="Set null character"),
        Argument("--no-header", default=False, help="Do not prepend header"),
    ]

    def run(self, args):
        args.delimiter = unescape_string(args.delimiter)
        if args.archive:
            if not args.output_file:
                log.write("An output file must be specified when using archive encoding")
                return
            args.no_header = False
        elif args.json:
            args.no_header = True

        register_graceful_shutdown_signal()
        with TeradataBulkExport(log_level=args.log_level, config=args.conf, key_file=args.key,
                dsn=args.dsn) as export:
            export.query = args.query
            start_time = time.time()
            if args.archive:
                with Writer(args.output_file, 'wb', use_gzip=args.gzip) as out:
                    i = 0
                    for n in export.to_archive(out):
                        i += n
                        log.info("\rExport", "Processed {} rows".format(int(round(i, -5))), console=True)
                    log.info("\rExport", "Processed {} rows".format(i))
            else:
                with Writer(args.output_file, use_gzip=args.gzip) as out:
                    export.options("output", out.name, 4)
                    export.options("delimiter", escape_string(args.delimiter), 2)
                    export.options("null", args.null, 3)
                    if args.json:
                        exportfn = export.to_json
                        export.options("encoding", "json", 5)
                    else:
                        exportfn = lambda: export.to_str(args.delimiter, args.null)
                        export.options("encoding", "str", 5)
                    export._initiate()
                    if out.is_stdout:
                        log.info(colors.green(colors.bold("-"*32)))
                    if not args.no_header:
                        out.writen(args.delimiter.join(export.columns.names))
                    i = 0
                    for i, row in enumerate(exportfn(), 1):
                        if i % 100000 == 0 and args.output_file:
                            log.info("\rExport", "Processed {} rows".format(i), console=True)
                        out.writen(row)
                    if args.output_file:
                        log.info("\rExport", "Processed {} rows".format(i))
                    if out.is_stdout:
                        log.info(colors.green(colors.bold("-"*32)))
            total_time = time.time() - start_time
            log.info("Results", "{} rows in {}".format(i, readable_time(total_time - export.idle_time)))
            log.info("Results", "Time spent idle: {}".format(readable_time(export.idle_time)))
            log.info("Results", "Total time spent: {}".format(readable_time(total_time)))


class ExternalCommand(Command):
    name = "external"
    description = "Run external commands"
    usage = "giraffez external <command> [options]"
    help = "Run external shell commands from a YAML job file."

    arguments = [
        Argument("command", help="External shell command to run"),
        Argument("--silent", default=False),
        Argument("--shell", default=False),
        Argument("--panic", default=False),
    ]

    def run(self, args):
        stdout, stderr = None, None
        if args.silent:
            log.level = SILENCE
            stdout, stderr = subprocess.PIPE, subprocess.PIPE
        parsed_command = args.command if args.shell else shlex.split(args.command)
        try:
            return_code = subprocess.call(parsed_command, stdout=stdout, stderr=stderr, shell=args.shell)
        except OSError:
            return_code = 2
        log.info("Finished with return code: {}".format(return_code))
        if args.panic and return_code != 0:
            error = GiraffeError("Command '{}' finished with exit code {} and panic is set, aborting...".format(args.command, return_code))
            error.code = return_code
            raise error
        return return_code


class FmtCommand(Command):
    name = "fmt"
    description = "Formatting tools"
    usage = "giraffez fmt <input_file> [options]"
    help = "Teradata format helper"

    arguments = [
        Argument("input_file", help="Input to be transformed"),
        Argument("-r", "--regex", help="search and replace (regex)"),
        Argument("--head", const=9, nargs="?", type=int, help="Only output N rows"),
        Argument("--header", default=False, help="Output table header"),
        Argument("--count", default=False, help="Count the table")
    ]

    @timer
    def run(self, args):
        with Reader(args.input_file) as f:
            if args.count:
                log.verbose("Debug[1]", "File type: ", f)
                if isinstance(f, ArchiveFileReader):
                    columns = f.columns
                else:
                    columns = f.header
                for column in columns:
                    log.verbose("Debug[1]", repr(column))
                i = 0
                for i, line in enumerate(f, 1):
                    pass
                log.info("Lines: ", i)
                return
            elif args.head:
                for i, line in enumerate(f, 1):
                    if args.head < i:
                        break
                    sys.stdout.write(line)
                return
        with io.open(args.input_file) as f:
            if args.header:
                print(f.header)
            if args.regex:
                parts = args.regex.split('/')
                if len(parts) < 3 or parts[0] != "s":
                    raise GiraffeError("Regex pattern '{}' not recoginzed.".format(args.regex))
                pattern = re.compile(parts[1])
                processor = lambda line: pattern.sub(parts[2], line)
            i = 0
            for i, line in enumerate(f, 1):
                if args.head and args.head < i:
                    break
                sys.stdout.write(processor(line))


class InsertCommand(Command):
    name = "insert"
    description = "Insert data into Teradata (CLIv2)"
    usage = "giraffez insert <input_file> <table> [options]"
    help = "Insert data to Teradata table (CLIv2)"

    arguments = [
        Argument("input_file", help="Input file"),
        Argument("table", help="Name of table"),
        Argument("-d", "--delimiter", default=None, help="Text delimiter"),
        Argument("-n", "--null", default='NULL', help="Set null character"),
        Argument("--quote-char", default='"', help=("One-character string used to quote fields "
            "containing special characters")),
        Argument("--parse-dates", default=False, help=("Enable automatic conversion of "
            "dates/timestamps")),
        Argument("-p", "--no-panic", default=False, help="Prevents normal panic behavior on error")
    ]

    def run(self, args):
        if not file_exists(args.input_file):
            raise FileNotFound("File '{}' does not exist.".format(args.input_file))

        if FileReader.check_length(args.input_file, MLOAD_THRESHOLD):
            show_warning(("USING LOAD TO INSERT MORE THAN {} ROWS IS NOT RECOMMENDED - USE MLOAD "
                "INSTEAD!").format(MLOAD_THRESHOLD), UserWarning)

        if args.delimiter is None:
            args.delimiter = file_delimiter(args.input_file)

        register_graceful_shutdown_signal()
        with TeradataCmd(log_level=args.log_level, config=args.conf, key_file=args.key,
                dsn=args.dsn, panic=args.no_panic) as load:
            load.options("source", args.input_file, 0)
            load.options("output", args.table, 1)
            load.options("null", args.null, 5)
            start_time = time.time()
            result = load.from_file(args.table, args.input_file, null=args.null, delimiter=args.delimiter,
                parse_dates=args.parse_dates, quotechar=args.quote_char)
            log.info("Results", "{} errors; {} rows in {}".format(result['errors'], result['count'], readable_time(time.time() - start_time)))


class LoadCommand(Command):
    name = "load"
    description = "Load data into Teradata (MultiLoad)"
    usage = "giraffez load <input_file> <table> [options]"
    help = "Load data to existing Teradata table (MultiLoad)"

    arguments = [
        Argument("input_file", help="Input file"),
        Argument("table", help="Name of table"),
        Argument("-d", "--delimiter", default=None, help="Text delimiter [default: (inferred from header)]"),
        Argument("-n", "--null", default='NULL', help="Set null character"),
        Argument("-y", "--drop-all", default=False, help="Drop tables"),
        Argument("--quote-char", default='"', help="One-character string used to quote fields containing special characters"),
        Argument("--coerce-floats", default=False, help="Allow floating-point numbers to be converted to fixed-precision numbers."),
        Argument("--parse-dates", default=False, help=("Enable automatic conversion of "
            "dates/timestamps")),
    ]

    def run(self, args):
        if args.delimiter is not None:
            args.delimiter = unescape_string(args.delimiter)
        if not file_exists(args.input_file):
            raise GiraffeError("File '{}' does not exist.".format(args.input_file))

        if not FileReader.check_length(args.input_file, MLOAD_THRESHOLD):
            show_warning(("USING MLOAD TO INSERT LESS THAN {} ROWS IS NOT RECOMMENDED - USE LOAD "
                "INSTEAD!").format(MLOAD_THRESHOLD), UserWarning)

        register_graceful_shutdown_signal()
        with TeradataBulkLoad(log_level=args.log_level, config=args.conf, key_file=args.key,
                dsn=args.dsn) as load:
            load.coerce_floats = args.coerce_floats
            load.table = args.table
            if args.drop_all is True:
                load.cleanup()
            else:
                existing_tables = list(filter(lambda x: load.mload.exists(x), load.tables))
                if len(existing_tables) > 0:
                    log.info("BulkLoad", "Previous work tables found:")
                    for i, table in enumerate(existing_tables, 1):
                        log.info('  {}: "{}"'.format(i, table))
                    if prompt_bool("Do you want to drop these tables?", default=True):
                        load.cleanup()
                    else:
                        log.fatal("Cannot continue without dropping previous job tables. Exiting ...")
            log.info("BulkLoad", "Executing ...")
            log.info('  source      => "{}"'.format(args.input_file))
            log.info('  output      => "{}"'.format(args.table))
            start_time = time.time()
            exit_code = load.from_file(args.input_file, delimiter=args.delimiter, null=args.null, quotechar=args.quote_char, parse_dates=args.parse_dates)
            total_time = time.time() - start_time
            log.info("BulkLoad", "Teradata PT request finished with exit code '{}'".format(exit_code))
            log.info("Results", "...")
            log.info('  Successful   => "{}"'.format(load.applied_count))
            log.info('  Unsuccessful => "{}"'.format(load.error_count))
            log.info("Results", "{} rows in {}".format(load.total_count, readable_time(total_time - load.idle_time)))
            log.info("Results", "Time spent idle: {}".format(readable_time(load.idle_time)))
            log.info("Results", "Total time spent: {}".format(readable_time(total_time)))
            if exit_code != 0:
                with TeradataCmd(log_level=args.log_level, config=args.conf, key_file=args.key,
                        dsn=args.dsn, silent=True) as cmd:
                    result = list(cmd.execute("select a.errorcode, a.errorfield, count(*) over (partition by a.errorcode, a.errorfield) as errorcount, b.errortext, min(substr(a.hostdata, 0, 30000)) as hostdata from {}_e1 a join dbc.errormsgs b on a.errorcode = b.errorcode qualify row_number() over (partition by a.errorcode, a.errorfield order by a.errorcode asc)=1 group by 1,2,4".format(args.table)))
                    if len(result) == 0:
                        log.info("No error information available.")
                        return
                    error_table = [("Error Code", "Field", "Error Count", "Error Description")]
                    samples = []
                    for row in result:
                        error_table.append((row[0], row[1], row[2], row[3]))
                        samples.append(repr(row[4]))
                    #log.level = args.log_level
                    log.info("\n{}\n".format(format_table(error_table)))
                    log.info("Sample Row Error Data:")
                    log.info("\t{}".format("\n".join(samples)))
                    #log.level = SILENCE


class RunCommand(Command):
    name = "run"
    description = "Run YAML job scripts"
    usage = "giraffez run <job_file> [options]"
    help = "Run giraffez job file (YAML)"

    arguments = [
        Argument("job_file", help="YAML script file"),
        Argument("definitions", default=[], nargs="*", help="Define variables used in the target job file. Arguments should be of the format <name>=<value>, and used in job files as {<name>}.')"),
    ]

    def run(self, args):
        params = {}
        for d in args.definitions:
            d = d.strip()
            if "=" in d:
                l, r = d.split("=")
                params[l] = r
            else:
                log.fatal("Invalid definition '{}': arguments must be of the format 'name=value'.".format(d))
        try:
            with open(args.job_file, "rb") as f:
                data = f.read()
        except IOError:
            log.fatal("Job file '{}' could not be read.".format(args.job_file))
        try:
            data = data.format(**params)
        except KeyError as k:
            log.fatal("Job file '{}' contained unknown format variable {}".format(args.job_file, k))
        try:
            job_config = yaml.load(data)
        except ValueError:
            job_config = False
        commands = {
            "cmd": CmdCommand,
            "config": ConfigCommand,
            "export": ExportCommand,
            "external": ExternalCommand,
            "fmt": FmtCommand,
            "load": LoadCommand,
            "run": RunCommand,
            "shell": ShellCommand,
            "secret": SecretCommand,
        }
        register_graceful_shutdown_signal()
        for job in job_config:
            job_type = job.get("type")
            if not job_type:
                log.fatal("No job")
            command = commands.get(job_type)
            if not command:
                log.fatal("No comm")
            c = command()
            default_args = c.parse_known_args()[0]
            all_args = argparse.Namespace()
            for (k, v) in default_args._get_kwargs():
                setattr(all_args, k, v)
            for k, v in job.get("settings", {}).items():
                setattr(all_args, k, v)
            c.run(all_args)


class SecretCommand(Command):
    name = "secret"
    description = "giraffez encrypted storage"
    usage = "giraffez secret <key>"
    help = "Retrieve encrypted settings"


    arguments = [
        Argument("key_value"),
        Argument("-n", default=False),
    ]

    def run(self, args):
        with Config(conf=args.conf, key_file=args.key) as c:
            if not args.key_value.startswith("secure.") and \
                    not args.key_value.startswith("connections."):
                key = "secure.{}".format(args.key_value)
            else:
                key = args.key_value
            value = c.get_value(key)
            if isinstance(value, basestring):
                sys.stdout.write(value)
                if args.n:
                    sys.stdout.write("\n")
            else:
                sys.stderr.write("Key '{}' not found\n".format(key))
                sys.exit(1)


class ShellCommand(Command):
    name = "shell"
    description = "Interactive Teradata shell"
    usage = "giraffez shell [options]"
    help = "Interactive Teradata shell"

    arguments = [
        Argument("-t", "--table-output", default=False, help="Print in table format"),
    ]

    def run(self, args):
        with TeradataCmd(log_level=SILENCE, config=args.conf, key_file=args.key, dsn=args.dsn) as cmd:
            shell = GiraffeShell(cmd)
            shell.table_output = args.table_output
            try:
                shell.cmdloop()
            except KeyboardInterrupt:
                log.fatal()



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap

from .constants import *
from .errors import *

from .logging import log

from ._compat import *


def escape_quotes(s):
    return s.replace("'", "''")

def format_indent(body, indent=" ", initial=""):
    wrapper = textwrap.TextWrapper(initial_indent=initial,
        subsequent_indent=indent)
    return wrapper.fill(body)

def format_table(lines, header=True):
    widths = [0] * len(lines[0])
    for line in lines:
        widths = [max(a, len(str(b))) for a, b in zip(widths, line)]
    output = []
    for i, line in enumerate(lines):
        output.append(" | ".join([str(field).ljust(w) for w, field in zip(widths, line)]))
        if i == 0 and header:
            output.append("-"*len(output[0]))
    return "\n".join(output)

def safe_name(s):
    return s.replace(" ", "_").lower()

def quote_string(s, quote_character="'"):
    return "{0}{1}{0}".format(quote_character, s)

def replace_cr(s):
    if not isinstance(s, str):
        return s
    return s.strip().replace('\r', '\n')

def truncate(s, n=7, c=70, placeholder="..."):
    if not isinstance(s, basestring):
        raise GiraffeError("Cannot truncate non-string value")
    lines = s.split('\n')
    line_count = len(lines)
    char_count = len(s)
    if char_count < (n * c) and line_count < n:
        return s
    if char_count > (n * c):
        s = s[0:n*c]
        lines = s.split('\n')
    lines = "\n".join(lines[0:n])
    return "{} {}".format(lines, placeholder)



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import datetime
import struct
try:
    import ujson as json
except ImportError:
    import json

from ._teradata import Encoder
from .constants import *
from .errors import *

from .fmt import escape_quotes, quote_string, replace_cr
from .logging import log
from .types import Columns, Date, Decimal, Time, Timestamp
from .utils import pipeline

from ._compat import *



class TeradataEncoder(object):
    """
    The class wrapping the Teradata C encoder.

    Exposed under the alias :class:`giraffez.Encoder`.

    :param list or :class:`giraffez.Columns` columns: Columns used for encoding.
    :param int encoding: Constant value representing the settings used by underlying C encoder.
    """
    def __init__(self, columns=[], encoding=None):
        self.encoding = ENCODER_SETTINGS_DEFAULT
        self._columns = columns
        self._delimiter = '|'
        self._null = None
        self.encoder = Encoder(columns)
        if encoding is not None:
            self |= encoding

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, value):
        if not isinstance(value, Columns):
            value = Columns(value)
        self._columns = value
        self.encoder.set_columns(self._columns)

    @classmethod
    def count(self, data):
        return Encoder.count_rows(data)

    @property
    def delimiter(self):
        return self._delimiter

    @delimiter.setter
    def delimiter(self, value):
        self._delimiter = value
        self.encoder.set_delimiter(self._delimtier)

    @property
    def null(self):
        return self._null

    @null.setter
    def null(self, value):
        self._null = value
        self.encoder.set_null(self._null)

    def parse_header(self, data):
        return self.encoder.unpack_stmt_info(data)

    def read(self, data):
        return self.encoder.unpack_row(data)

    def readbuffer(self, data):
        return self.encoder.unpack_rows(data)

    def serialize(self, data):
        return self.encoder.pack_row(data)

    def setdefaults(self):
        self.encoding = ENCODER_SETTINGS_DEFAULT
        self.encoder.set_encoding(self.encoding)

    def __or__(self, other):
        if other & ROW_RETURN_MASK:
            self.encoding = self.encoding & ~ROW_RETURN_MASK | other
        if other & DATETIME_RETURN_MASK:
            self.encoding = self.encoding & ~DATETIME_RETURN_MASK | other
        if other & DECIMAL_RETURN_MASK:
            self.encoding = self.encoding & ~DECIMAL_RETURN_MASK | other
        self.encoder.set_encoding(self.encoding)
        self.encoder.set_delimiter(self._delimiter)
        self.encoder.set_null(self._null)
        return self

    def __str__(self):
        return "{}|{}|{}".format(
            encoder_settings_map.get(self.encoding & ROW_RETURN_MASK, "Unknown"),
            encoder_settings_map.get(self.encoding & DATETIME_RETURN_MASK, "Unknown"),
            encoder_settings_map.get(self.encoding & DECIMAL_RETURN_MASK, "Unknown"),
        )


class Handler(object):
    """
    Base class for creating data handlers.  This class can be subclassed
    to define a list of functions that run on data that matches the
    Teradata data types provided.  This is useful when dealing with data
    that may not be in the appropriate Python type to be encoded.
    """
    handlers = []

    def __init__(self, columns):
        if not isinstance(columns, Columns):
            raise GiraffeEncodeError("Type '{}' must be type giraffez.types.Columns".format(type(columns)))
        self.columns = columns
        self._columns = []
        for index, column in enumerate(self.columns):
            for cond, fn in self.handlers:
                if column.type in cond:
                    self._columns.append((fn, index))

    def __call__(self, data):
        for fn, index in self._columns:
            if data[index] is not None:
                data[index] = fn(data[index])
        return data


class CharHandler(Handler):
    handlers = [
        (CHAR_TYPES | VAR_TYPES, replace_cr),
    ]


class DateHandler(Handler):
    handlers = [
        (DATE_TYPES, Date.from_integer),
        (TIME_TYPES, Time.from_string),
        (TIMESTAMP_TYPES, Timestamp.from_string)
    ]


class DecimalHandler(Handler):
    handlers = [
        (DECIMAL_TYPES, Decimal),
    ]


class FloatHandler(Handler):
    handlers = [
        (DECIMAL_TYPES, pipeline([float, str])),
    ]


def check_input(columns, items):
    if len(items) != len(columns):
        error_message = "Columns list contains {} columns and {} items provided.".format(len(columns), len(items))
        raise GiraffeEncodeError("{}: \n\t{!r}".format(error_message, items))

def null_handler(null=None):
    nulls = {"", None, null}
    def _handler(items):
        return [item if item not in nulls else None for item in items]
    return _handler

def dict_to_json(record):
    try:
        return json.dumps(record, ensure_ascii=False)
    except UnicodeDecodeError as error:
        log.debug("Debug[2]", "UnicodeDecodeError: {}".format(error.message))

def python_to_sql(table_name, columns, date_conversion=True):
    if date_conversion:
        def convert_time(s):
            value = Time.from_string(s)
            if value is None:
                raise GiraffeEncodeError("Cannot convert time '{}'".format(s))
            return value.to_string()

        def convert_date(s):
            value = Date.from_string(s)
            if value is None:
                raise GiraffeEncodeError("Cannot convert date '{}'".format(s))
            return value.to_string()

        def convert_datetime(s, length):
            if isinstance(s, datetime.date):
                value = Timestamp.from_datetime(s)
            else:
                value = Timestamp.from_string(s)
            if value is None:
                raise GiraffeEncodeError("Cannot convert datetime '{}'".format(s))
            return value.to_string(length)
    else:
        convert_time = lambda a: a
        convert_date = lambda a: a
        convert_datetime = lambda a, b: a
    def _encoder(items):
        check_input(columns, items)
        values = []
        for item, column in zip(items, columns):
            if item is None:
                continue
            if column.type in CHAR_TYPES | VAR_TYPES:
                value = quote_string(escape_quotes(item))
            elif column.type in DATE_TYPES:
                if isinstance(item, datetime.date):
                    value = str(Date.from_datetime(item).to_integer())
                else:
                    value = quote_string(convert_date(escape_quotes(item)))
            elif column.type in TIME_TYPES:
                value = quote_string(convert_time(item))
            elif column.type in TIMESTAMP_TYPES:
                value = quote_string(convert_datetime(item, column.length))
            else:
                value = str(item)
            values.append((column.name, value))
        return "ins into {} ({}) values ({});".format(table_name,
            ",".join(quote_string(x[0], '"') for x in values),
            ",".join(x[1] for x in values))
    return _encoder



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import struct
import threading

from ._teradatapt import EncoderError, InvalidCredentialsError, MLoad, TeradataError as TeradataPTError

from .constants import *
from .errors import *

from .cmd import TeradataCmd
from .connection import Connection, Context
from .encoders import DateHandler, null_handler
from .fmt import format_table
from .io import ArchiveFileReader, CSVReader, FileReader, JSONReader, Reader
from .logging import log
from .utils import get_version_info, pipeline, suppress_context

from ._compat import *


__all__ = ['BulkLoad']


error_table_count = """
select a.errorcode as code, a.errorfield as field, b.errortext as text, count(*) over (partition by a.errorcode, a.errorfield) as errcount
from {0}_e1 a
join dbc.errormsgs b
on a.errorcode = b.errorcode
group by 1,2,3;
"""

error_table_sample = """
select
a.errorcode as code,
a.errorfield as field,
min(substr(a.hostdata, 0, 30000)) as hostdata
from {0}_e1 a
qualify row_number() over (partition by a.errorcode, a.errorfield order by a.errorcode asc)=1
group by 1,2;
"""


class TeradataBulkLoad(Connection):
    """
    The class for using the TPT API's UPDATE (MLoad) driver to insert a
    large (> ~100k rows) amount of data into an existing Teradata table.

    Exposed under the alias :class:`giraffez.BulkLoad`.

    :param str table: The name of the target table for loading.
    :param str host: Omit to read from :code:`~/.girafferc` configuration file.
    :param str username: Omit to read from :code:`~/.girafferc` configuration file.
    :param str password: Omit to read from :code:`~/.girafferc` configuration file.
    :param int log_level: Specify the desired level of output from the job.
        Possible values are :code:`giraffez.SILENCE`, :code:`giraffez.INFO` (default),
        :code:`giraffez.VERBOSE`, and :code:`giraffez.DEBUG`
    :param str config: Specify an alternate configuration file to be read from,
        when previous paramaters are omitted.
    :param str key_file: Specify an alternate key file to use for configuration decryption
    :param string dsn: Specify a connection name from the configuration file to be
        used, in place of the default.
    :param bool protect: If authentication with Teradata fails and :code:`protect` is :code:`True`,
        locks the connection used in the configuration file. This can be unlocked using the
        command :code:`giraffez config --unlock <connection>`, changing the connection password,
        or via the :meth:`~giraffez.config.Config.unlock_connection` method.
    :param bool coerce_floats: Coerce Teradata decimal types into Python floats
    :param bool cleanup: Attempt to cleanup all work tables
        when context exits.
    :param bool print_error_table: Prints a user-friendly version of the mload
        error table to stderr.
    :raises `giraffez.errors.InvalidCredentialsError`: if the supplied credentials are incorrect
    :raises `giraffez.TeradataPTError`: if the connection cannot be established

    If the target table is currently under an MLoad lock (such as if the
    previous operation failed), a :code:`release mload` statement will be
    executed on the table, and the load job will be re-attempted.

    Meant to be used, where possible, with python's :code:`with` context handler
    to guarantee that connections will be closed gracefully when operation
    is complete.
    """
    checkpoint_interval = 50000

    def __init__(self, table=None, host=None, username=None, password=None,
            log_level=INFO, config=None, key_file=None, dsn=None, protect=False,
            coerce_floats=False, cleanup=False, print_error_table=False):
        super(TeradataBulkLoad, self).__init__(host, username, password, log_level, config, key_file,
            dsn, protect)
        # Attributes used with property getter/setters
        self._columns = None
        self._table_name = None
        self.initiated = False
        self.finished = False
        self.coerce_floats = coerce_floats
        self.perform_cleanup = cleanup
        self.exit_code = None
        self.applied_count = 0
        self.error_count = 0
        #: The amount of time spent in idle (waiting for server)
        self.idle_time = 0
        #: Prints the error table when there is an issue, good for troubleshooting jobs
        self.print_error_table = print_error_table
        self.preprocessor = lambda s: s
        if table is not None:
            self.table = table

    def checkpoint(self):
        """
        Execute a checkpoint while loading rows. Called automatically
        when loading from a file. Updates the exit code of the driver to
        reflect errors.
        """
        return self.mload.checkpoint()

    def cleanup(self):
        """
        Drops any existing work tables, as returned by
        :meth:`~giraffez.load.TeradataBulkLoad.tables`.

        :raises `giraffez.TeradataPTError`: if a Teradata error ocurred
        """
        threads = []
        for i, table in enumerate(filter(lambda x: self.mload.exists(x), self.tables)):
            log.info("BulkLoad", "Dropping table '{}'...".format(table))
            t = threading.Thread(target=self.mload.drop_table, args=(table,))
            threads.append(t)
            t.start()
        for t in threads:
            t.join()

    @property
    def columns(self):
        """
        The list of columns in use.

        :getter: Return the list of columns in use.
        :setter: Set the columns to be loaded into, as well as their order. If
            loading from a file, these will be determined from the file header.
            Not necessary if you are loading into all columns, in the original
            order. The value must be a :code:`list` of names in the order that
            the fields of data will be presented in each row.

            Raises :class:`~giraffez.errors.GiraffeError` if :code:`field_names`
            is not a :code:`list`.

            Raises :class:`~giraffez.errors.GiraffeError` if the target table
            has not been set.
        :type: :class:`~giraffez.types.Columns`
        """
        return self._columns

    @columns.setter
    def columns(self, field_names):
        if not isinstance(field_names, list):
            raise GiraffeError("Must set .columns property as type <List>")
        self._columns = field_names

    def finish(self):
        """
        Finishes the load job. Called automatically when the connection closes.

        :return: The exit code returned when applying rows to the table
        """
        if self.finished:
            return self.exit_code
        checkpoint_status = self.checkpoint()
        self.exit_code = self._exit_code()
        if self.exit_code != 0:
            raise TeradataPTError("BulkLoad job finished with return code '{}'".format(self.exit_code))
        # TODO(chris): should this happen every time?
        if self.applied_count > 0:
            self._end_acquisition()
            self._apply_rows()
        self.exit_code = self._exit_code()
        if self.exit_code != 0:
            raise TeradataPTError("BulkLoad job finished with return code '{}'".format(self.exit_code))
        self.finished = True
        return self.exit_code

    def from_file(self, filename, table=None, delimiter='|', null='NULL',
            panic=True, quotechar='"', parse_dates=False):
        """
        Load from a file into the target table, handling each step of the
        load process.

        Can load from text files, and properly formatted giraffez archive
        files. In both cases, if Gzip compression is detected the file will be
        decompressed while reading and handled appropriately. The encoding is
        determined automatically by the contents of the file.

        It is not necessary to set the columns in use prior to loading from a file.
        In the case of a text file, the header is used to determine column names
        and their order. Valid delimiters include '|', ',', and '\\t' (tab). When
        loading an archive file, the column information is decoded alongside the data.

        :param str filename: The location of the file to be loaded
        :param str table: The name of the target table, if it was not specified
            to the constructor for the isntance
        :param str null: The string that indicates a null value in the rows being
            inserted from a file. Defaults to 'NULL'
        :param str delimiter: When loading a file, indicates that fields are
            separated by this delimiter. Defaults to :code:`None`, which causes the
            delimiter to be determined from the header of the file. In most
            cases, this behavior is sufficient
        :param str quotechar: The character used to quote fields containing special characters,
            like the delimiter.
        :param bool panic: If :code:`True`, when an error is encountered it will be
            raised. Otherwise, the error will be logged and :code:`self.error_count`
            is incremented.
        :return: The output of the call to
            :meth:`~giraffez.load.TeradataBulkLoad.finish`
        :raises `giraffez.errors.GiraffeError`: if table was not set and :code:`table`
            is :code:`None`, or if a Teradata error ocurred while retrieving table info.
        :raises `giraffez.errors.GiraffeEncodeError`: if :code:`panic` is :code:`True` and there
            are format errors in the row values.
        """
        if not self.table:
            if not table:
                raise GiraffeError("Table must be set or specified to load a file.")
            self.table = table
        if not isinstance(null, basestring):
            raise GiraffeError("Expected 'null' to be str, received {}".format(type(null)))
        with Reader(filename, delimiter=delimiter, quotechar=quotechar) as f:
            if not isinstance(f.delimiter, basestring):
                raise GiraffeError("Expected 'delimiter' to be str, received {}".format(type(delimiter)))
            if isinstance(f, ArchiveFileReader):
                self.mload.set_encoding(ROW_ENCODING_RAW)
                self.columns = f.header
                self.preprocessor = lambda s: s
            if parse_dates:
                self.preprocessor = DateHandler(self.columns)
            self._initiate()
            self.mload.set_null(null)
            self.mload.set_delimiter(delimiter)
            i = 0
            for i, line in enumerate(f, 1):
                self.put(line, panic=panic)
                if i % self.checkpoint_interval == 1:
                    log.info("\rBulkLoad", "Processed {} rows".format(i), console=True)
                    checkpoint_status = self.checkpoint()
                    self.exit_code = self._exit_code()
                    if self.exit_code != 0:
                        return self.exit_code
            log.info("\rBulkLoad", "Processed {} rows".format(i))
            return self.finish()

    def put(self, items, panic=True):
        """
        Load a single row into the target table.

        :param list items: A list of values in the row corresponding to the
            fields specified by :code:`self.columns`
        :param bool panic: If :code:`True`, when an error is encountered it will be
            raised. Otherwise, the error will be logged and :code:`self.error_count`
            is incremented.
        :raises `giraffez.errors.GiraffeEncodeError`: if :code:`panic` is :code:`True` and there
            are format errors in the row values.
        :raises `giraffez.errors.GiraffeError`: if table name is not set.
        :raises `giraffez.TeradataPTError`: if there is a problem
            connecting to Teradata.
        """
        if not self.initiated:
            self._initiate()
        try:
            row_status = self.mload.put_row(self.preprocessor(items))
            self.applied_count += 1
        except (TeradataPTError, EncoderError) as error:
            self.error_count += 1
            if panic:
                raise error
            log.info("BulkLoad", error)

    def read_error_table(self):
        with TeradataCmd(log_level=log.level, config=self.config, key_file=self.key_file,
                dsn=self.dsn, silent=True) as cmd:
            result = list(cmd.execute((error_table_count + error_table_sample).format(self.table)))
            if len(result) == 0:
                log.info("No error information available.")
                return
            return result

    def release(self):
        """
        Attempt release of target mload table.

        :raises `giraffez.errors.GiraffeError`: if table was not set by
            the constructor, the :code:`TeradataBulkLoad.table`, or
            :meth:`~giraffez.load.TeradataBulkLoad.from_file`.
        """
        if self.table is None:
            raise GiraffeError("Cannot release. Target table has not been set.")
        log.info("BulkLoad", "Attempting release for table {}".format(self.table))
        self.mload.release(self.table)

    @property
    def status(self):
        return self.mload.status()

    @property
    def tables(self):
        """
        The names of the work tables used for loading.

        :return: A list of four tables, each the name of the target table
            with the added suffixes, "_wt", "_log", "_e1", and "_e2"
        :raises `giraffez.errors.GiraffeError`: if table was not set by
            the constructor, the :code:`TeradataBulkLoad.table`, or
            :meth:`~giraffez.load.TeradataBulkLoad.from_file`.
        """
        if self.table is None:
            raise GiraffeError("Target table has not been set.")
        return [
            "{}_wt".format(self.table),
            "{}_log".format(self.table),
            "{}_e1".format(self.table),
            "{}_e2".format(self.table),
        ]

    @property
    def table(self):
        """
        The name of the target table.

        :getter: Returns the name of the target table, or :code:`None` if it
            has not been set.
        :setter: Set the name of the target table, if the table name
            was not given to the constructor of the
            :class:`~giraffez.load.TeradataBulkLoad` instance or
            :meth:`~giraffez.load.TeradataBulkLoad.from_file`. The value given
            must include all qualifiers such as database name.

            Raises :class:`~giraffez.errors.GiraffeError` if the MLoad connection has
            already been initiated, or the :class:`~giraffez.cmd.TeradataCmd` connection cannot
            be established.

            Raises :class:`~giraffez.TeradataPTError` if the column data could not be
            retrieved from Teradata
        :type: str
        """
        return self._table_name

    @table.setter
    def table(self, table_name):
        if self.initiated:
            raise GiraffeError("Cannot reuse BulkLoad context for more than one table")
        self._table_name = table_name

    @property
    def total_count(self):
        """
        The number of rows applied, plus the number of rows in error.
        """
        return self.applied_count + self.error_count

    def _apply_rows(self):
        log.info("BulkLoad", "Beginning apply phase ...")
        self.mload.apply_rows()
        self._update_apply_count()
        self._update_error_count()
        log.info("BulkLoad", "Apply phase ended.")

    def _close(self, exc=None):
        if not self.initiated:
            return
        if not exc:
            try:
                self.finish()
            except TeradataPTError as error:
                self._close(error)
                if self.print_error_table:
                    for row in self.read_error_table():
                        print(row)
                raise error
        log.info("BulkLoad", "Closing Teradata PT connection ...")
        self.mload.close()
        log.info("BulkLoad", "Teradata PT request complete.")

    def _connect(self, host, username, password):
        self.mload = MLoad(host, username, password)
        title, version = get_version_info()
        query_band = "UTILITYNAME={};VERSION={};".format(title, version)
        self.mload.add_attribute(TD_QUERY_BAND_SESS_INFO, query_band)

    def _end_acquisition(self):
        log.info("BulkLoad", "Ending acquisition phase ...")
        self.mload.end_acquisition()
        log.info("BulkLoad", "Acquisition phase ended.")

    def _exit_code(self):
        data = self.mload.get_event(TD_Evt_ExitCode)
        if data is None:
            log.info("BulkLoad", "Update exit code failed.")
            return
        return struct.unpack("h", data)[0]

    def _initiate(self):
        if not self.table:
            raise GiraffeError("Table must be set prior to initiating.")
        if self.initiated:
            raise GiraffeError("Already initiated connection.")
        try:
            if self.perform_cleanup:
                self.cleanup()
            elif any(filter(lambda x: self.mload.exists(x), self.tables)):
                raise GiraffeError("Cannot continue without dropping previous job tables. Exiting ...")
            log.info("BulkLoad", "Initiating Teradata PT request (awaiting server)  ...")
            start_time = time.time()
            self.mload.initiate(self.table, self.columns)
            self.idle_time = time.time() - start_time
        except InvalidCredentialsError as error:
            if args.protect:
                Config.lock_connection(args.conf, args.dsn, args.key)
            raise error
        self.mload.set_null(None)
        log.info("BulkLoad", "Teradata PT request accepted.")
        self._columns = self.mload.columns()
        self.initiated = True

    def _update_apply_count(self):
        data = self.mload.get_event(TD_Evt_RowCounts64)
        if data is None:
            log.info("BulkLoad", "Update apply row count failed.")
            return
        recv, sent, applied = struct.unpack("QQQ", data)
        log.debug("Debug[2]", "Event[RowCounts64]: r:{}, s:{}, a:{}".format(recv, sent, applied))
        self.applied_count = applied

    def _update_error_count(self):
        data = self.mload.get_event(TD_Evt_ErrorTable2, 1)
        if data is None:
            log.info("BulkLoad", "Update error row count failed.")
            return
        count = struct.unpack("I", data)[0]
        log.debug("Debug[2]", "Event[ErrorTable2]: c:{}".format(count))
        self.error_count = count


class BulkLoad(Context):
    __instance__ = TeradataBulkLoad



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class GiraffeError(Exception):
    """
    Baseclass for all giraffez errors.
    """

class GiraffeTypeError(GiraffeError):
    """
    Baseclass for all giraffez type errors.
    """

class GiraffeEncodeError(GiraffeError):
    """
    Raised when unable to encode the provided object.
    """

class ConfigurationError(GiraffeError): 
    """
    For use with configuration file handling. 
    """

class FileNotFound(GiraffeError): 
    """
    Raised when file does not exist.
    """

class ConfigNotFound(ConfigurationError, FileNotFound):
    """
    Raised when the specified configuration file does not exist.
    """

class KeyNotFound(ConfigurationError, FileNotFound):
    """
    Raised when the specified configuration file does not exist.
    """

class ConfigReadOnly(ConfigurationError):
    """
    Raised when a write is attempted on a configuration file was opened
    in read mode.
    """
class InvalidCredentialsError(ConfigurationError):
    """
    Raised when connection credentials are incorrect.
    """

class ConnectionLock(ConfigurationError):
    """
    Raised when connection is locked by invalid attempts and the
    'protect' feature is being used.
    """

    def __init__(self, dsn):
        super(ConnectionLock, self).__init__(("Connection {0} is currently locked. please update "
            "credentials and run:\n\tgiraffez config --unlock {0}").format(dsn))



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import argparse

from .constants import *
from .errors import *

from .logging import log
from .utils import get_version_info, import_string

from ._compat import *


__all__ = ["Argument", "Command"]


class Argument(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.help = kwargs.get("help", "")
        self.nargs = kwargs.get("nargs", None)
        self.default = kwargs.get("default", None)
        self.action = kwargs.get("action", "store")
        self.type = kwargs.get("type", None)
        self.const = kwargs.get("const", None)
        self.version = kwargs.get("version", "")
        self.group = kwargs.get("group", None)
        self.dest = kwargs.get("dest", None)
        if self.default is False:
            self.action = "store_true"
        if self.default is True:
            self.action = "store_false"


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super(ArgumentParser, self).__init__(*args, **kwargs)
        for group in self._action_groups:
            if group.title == "optional arguments":
                group.title = "options"

    def add_or_get_group(self, group_name):
        for group in self._action_groups:
            if group_name == group.title:
                return group
        if group_name == "mutual":
            for group in self._mutually_exclusive_groups:
                return group
            group = self.add_mutually_exclusive_group(required=True)
            return group
        else:
            return self.add_argument_group(group_name)

    def error(self, message):
        if "too few arguments" not in message:
            sys.stderr.write("error: {}\n".format(message))
        self.print_help()
        sys.exit(2)


class Command(ArgumentParser):
    required_arguments = []
    optional_arguments = []

    name = None
    description = None
    usage = None
    help = None

    arguments = []
    global_arguments = []
    commands = []

    default = "run"

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(self.name, description=self.description, usage=self.usage,
            formatter_class=HelpFormatter, add_help=False, conflict_handler='resolve')
        self.has_subcommands = False
        self.register_args(self.arguments)

        for global_arg in self.global_arguments:
            global_arg.group = "global options"
        self.register_args(self.global_arguments)
        self.set_defaults(name=self.name)
        if self.default:
            if isinstance(self.default, basestring):
                self.set_defaults(run=getattr(self, self.default))
            else:
                self.set_defaults(run=self.default)
        self.set_defaults(required=self.required_arguments)
        self.register_commands()

    def add_command(self, parent):
        if self.has_subcommands is False:
            self.sub_parser = self.add_subparsers(title="commands", dest="module_name",
                metavar=None)
            self.sub_parser._parser_class = ArgumentParser
            self.has_subcommands = True
        new_parser = self.sub_parser.add_parser(parent.prog, usage=parent.usage, help=parent.help,
            description=parent.description, parents=[parent], formatter_class=HelpFormatter,
            add_help=False, conflict_handler='resolve')

    def register_args(self, arguments):
        for argument in arguments:
            if argument.group:
                parser = self.add_or_get_group(argument.group)
            else:
                parser = self
            if argument.default:
                argument.help = "{} [default: '{}']".format(argument.help, argument.default)
            kwargs = {
                'help': argument.help,
                'default': argument.default,
                'action': argument.action
            }

            if argument.action == "store":
                kwargs['nargs'] = argument.nargs
                if argument.type:
                    kwargs['type'] = argument.type
                    kwargs['const'] = argument.const
            elif argument.action == "version":
                kwargs['version'] = " version ".join(get_version_info())
            if argument.dest:
                kwargs['dest'] = argument.dest
            parser.add_argument(
                *argument.args,
                **kwargs
            )

    def register_commands(self):
        for command in self.commands:
            if isinstance(command, Command):
                self.add_command(command)
            elif isinstance(command, basestring):
                try:
                    command = import_string(command)
                    command.global_arguments = self.global_arguments
                    c = command()
                    self.add_command(c)
                except ImportError as error:
                    log.info(error)
            else:
                command.global_arguments = self.global_arguments
                c = command()
                self.add_command(c)


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _format_action(self, action):
        parts = super(HelpFormatter, self)._format_action(action)

        # pop off subparser header line
        if isinstance(action, argparse._SubParsersAction):
            parts = "\n".join(parts.split('\n')[1:])
        return parts

    def _format_args(self, action, default_metavar):
        result = super(HelpFormatter, self)._format_args(action, default_metavar)
        return "<{}>".format(result.lower())



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import os
import sys
import time
import getpass
import functools
import itertools
import warnings

from .constants import *
from .errors import *

from .logging import log

from ._compat import *


__all__ = ['disable_warnings', 'get_version_info', 'import_string', 'pipeline', 'prompt',
    'prompt_bool', 'readable_time', 'register_graceful_shutdown_signal', 'show_warning',
    'suppress_context', 'timer']


def disable_warnings(category=UserWarning):
    warnings.simplefilter('ignore', category)

def get_version_info():
    from giraffez.__init__ import __title__, __version__
    return __title__, __version__

def import_string(import_name):
    import_name = str(import_name)
    if not import_name.startswith("giraffez."):
        import_name = "giraffez.{}".format(import_name)
    try:
        return __import__(import_name)
    except ImportError:
        try:
            module_name, obj_name = import_name.rsplit('.', 1)
        except AttributeError as error:
            raise ImportError(error)
        try:
            module = __import__(module_name, None, None, [obj_name])
        except ImportError:
            module = import_string(module_name)
        try:
            return getattr(module, obj_name)
        except AttributeError as error:
            raise ImportError(error)

def pipeline(funcs):
    def _pipeline(x):
        return reduce(lambda a, f: f(a), funcs, x)
    return _pipeline

def prompt(message, default=None, password=False):
    while True:
        if password:
            user_input = getpass.getpass(message)
        else:
            user_input = input(message)
        if user_input:
            return user_input 
        if default is not None:
            return default

def prompt_bool(message, default=False):
    yes_choices = ("y", "yes")
    no_choices = ("n", "no")

    if default:
        message = "{} [Y/n]: ".format(message)
        default_str = "Y"
    else:
        message = "{} [y/N]: ".format(message)
        default_str = "N"
    while True:
        user_input = prompt(message, default=default_str)
        if user_input.lower() in yes_choices:
            return True
        elif user_input.lower() in no_choices:
            return False
        else:
            return default

def readable_time(t):
    if t > 300:
        return "{}m".format(t//60)
    else:
        return "{:.3f}s".format(t)

def register_shutdown_signal():
    """
    Registers a simple shutdown handler using C signals.
    """
    from ._teradata import register_shutdown_signal as _register
    _register()

def register_graceful_shutdown_signal():
    """
    Registers graceful shutdown handler using C signals. The first
    SIGINT will attempt to close the Teradata connections before
    exiting, and the second will shutdown whether the connections
    have been closed or not.
    """
    from ._teradata import register_graceful_shutdown_signal as _register
    _register()

def show_warning(message, category):
    def send_warnings_to_log(message, category, filename, lineno, file=None, line=None):
        log.write("{}:{}: {}: {}".format(filename, lineno, category.__name__, message))
    warnings.showwarning = send_warnings_to_log
    warnings.warn(message, category)

def suppress_context(exc):
    """
    Used to suppress the error context for Python 3 errors when this is
    an undesired behavior. To ensure capability with Python 2, this
    must be used in place of `raise Exception from None`
    """
    exc.__context__ = None
    exc.__cause__ = None
    return exc

def timer(f):
    @functools.wraps(f)
    def wrapper(*args, **kwds):
        start = time.time()
        result = f(*args, **kwds)
        elapsed = time.time() - start
        sys.stderr.write("{} took {:.4f}s to finish\n".format(f.__name__, elapsed))
        return result
    return wrapper



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import struct

from ._teradatapt import Export

from .constants import *
from .errors import *

from ._teradatapt import InvalidCredentialsError
from .config import Config
from .connection import Connection, Context
from .encoders import dict_to_json, TeradataEncoder
from .fmt import truncate
from .logging import log
from .sql import parse_statement, remove_curly_quotes
from .utils import get_version_info, show_warning, suppress_context

from ._compat import *


__all__ = ['BulkExport']


class TeradataBulkExport(Connection):
    """
    The class for using the Teradata Parallel Transport API to quickly export
    large amounts of data.

    Exposed under the alias :class:`giraffez.BulkExport`.

    :param str query: Either SQL query to execute and return results of, **or** the
        name of a table to export in its entirety
    :param str host: Omit to read from :code:`~/.girafferc` configuration file.
    :param str username: Omit to read from :code:`~/.girafferc` configuration file.
    :param str password: Omit to read from :code:`~/.girafferc` configuration file.
    :param int log_level: Specify the desired level of output from the job.
        Possible values are :code:`giraffez.SILENCE` :code:`giraffez.INFO` (default),
        :code:`giraffez.VERBOSE` and :code:`giraffez.DEBUG`
    :param str config: Specify an alternate configuration file to be read from, when previous paramters are omitted.
    :param str key_file: Specify an alternate key file to use for configuration decryption
    :param string dsn: Specify a connection name from the configuration file to be
        used, in place of the default.
    :param bool protect: If authentication with Teradata fails and :code:`protect` is :code:`True` 
        locks the connection used in the configuration file. This can be unlocked using the
        command :code:`giraffez config --unlock <connection>` changing the connection password,
        or via the :meth:`~giraffez.config.Config.unlock_connection` method.
    :param bool coerce_floats: Coerce Teradata decimal types into Python floats
    :raises `giraffez.errors.InvalidCredentialsError`: if the supplied credentials are incorrect
    :raises `giraffez.TeradataError`: if the connection cannot be established

    Meant to be used, where possible, with Python's :code:`with` context handler
    to guarantee that connections will be closed gracefully when operation
    is complete:

    .. code-block:: python

       with giraffez.BulkExport('dbc.dbcinfo') as export:
           print(export.header)
           for row in export.to_list():
               print(row)
    """

    def __init__(self, query=None, host=None, username=None, password=None,
            log_level=INFO, config=None, key_file=None, dsn=None, protect=False,
            coerce_floats=True):
        super(TeradataBulkExport, self).__init__(host, username, password, log_level, config, key_file,
            dsn, protect)
        # Attributes used with property getter/setters
        self._query = None
        self.coerce_floats = coerce_floats
        self.initiated = False
        #: The amount of time spent in idle (waiting for server)
        self.idle_time = 0
        self.query = query

    @property
    def columns(self):
        """
        Retrieve columns if necessary, and return them.

        :rtype: :class:`~giraffez.types.Columns`
        """
        return self.export.columns()
    
    @property
    def query(self):
        """
        :return: The current query if it has been set, otherwise :code:`None`
        :rtype: str
        """
        return self._query

    @query.setter
    def query(self, query):
        """
        Set the query to be run and initiate the connection with Teradata.
        Only necessary if the query/table name was not specified as an argument
        to the constructor of the instance.

        :param str query: Valid SQL query to be executed
        """
        if query is None:
            return
        if log.level >= VERBOSE:
            self.options("query", query, 6)
        else:
            self.options("query", truncate(query), 6)
        statements = parse_statement(remove_curly_quotes(query))
        if not statements:
            raise GiraffeError("Unable to parse SQL statement")
        if len(statements) > 1:
            show_warning(("MORE THAN ONE STATEMENT RECEIVED, EXPORT OPERATIONS ALLOW ONE "
                "STATEMENT - ONLY THE FIRST STATEMENT WILL BE USED."), RuntimeWarning)
        statement = statements[0]
        log.debug("Debug[2]", "Statement (sanitized): {!r}".format(statement))
        if not (statement.startswith("select ") or statement.startswith("sel ")):
            statement = "select * from {}".format(statement)
        if statement == self.query:
            return
        else:
            self._query = statement
        self.initiated = False
        # Since CLIv2 is used in set_query (instead of relying on the
        # colums from the TPT Export driver) and set_query will always
        # happen before calls to initiate, set_query will always fail
        # with InvalidCredentialsError before initiate despite initiate
        # presumably failing after this point as well.
        try:
            self.export.set_query(statement)
        except InvalidCredentialsError as error:
            if args.protect:
                Config.lock_connection(args.conf, args.dsn, args.key)
            raise error

    def to_archive(self, writer):
        """
        Writes export archive files in the Giraffez archive format.
        This takes a `giraffez.io.Writer` and writes archive chunks to
        file until all rows for a given statement have been exhausted.

        .. code-block:: python

            with giraffez.BulkExport("database.table_name") as export:
                with giraffez.Writer("database.table_name.tar.gz", 'wb', use_gzip=True) as out:
                    for n in export.to_archive(out):
                        print("Rows: {}".format(n))

        :param `giraffez.io.Writer` writer: A writer handling the archive output

        :rtype: iterator (yields ``int``)
        """
        if 'b' not in writer.mode:
            raise GiraffeError("Archive writer must be in binary mode")
        writer.write(GIRAFFE_MAGIC)
        writer.write(self.columns.serialize())
        i = 0
        for n, chunk in enumerate(self._fetchall(ROW_ENCODING_RAW), 1):
            writer.write(chunk)
            yield TeradataEncoder.count(chunk)

    def to_dict(self):
        """
        Sets the current encoder output to Python `dict` and returns
        a row iterator.

        :rtype: iterator (yields ``dict``)
        """
        return self._fetchall(ROW_ENCODING_DICT)

    def to_json(self):
        """
        Sets the current encoder output to json encoded strings and
        returns a row iterator.

        :rtype: iterator (yields ``str``)
        """
        return self._fetchall(ROW_ENCODING_DICT, processor=dict_to_json)

    def to_list(self):
        """
        Sets the current encoder output to Python `list` and returns
        a row iterator.

        :rtype: iterator (yields ``list``)
        """
        return self._fetchall(ROW_ENCODING_LIST)

    def to_str(self, delimiter='|', null='NULL'):
        """
        Sets the current encoder output to Python `str` and returns
        a row iterator.

        :param str null: The string representation of null values
        :param str delimiter: The string delimiting values in the output
            string

        :rtype: iterator (yields ``str``)
        """
        self.export.set_null(null)
        self.export.set_delimiter(delimiter)
        self.options("delimiter", escape_string(delimiter), 2)
        self.options("null", null, 3)
        return self._fetchall(ENCODER_SETTINGS_STRING, coerce_floats=False)

    def _close(self, exc=None):
        log.info("Export", "Closing Teradata PT connection ...")
        self.export.close()
        log.info("Export", "Teradata PT request complete.")

    def _connect(self, host, username, password):
        self.export = Export(host, username, password)
        self.export.add_attribute(TD_QUERY_BAND_SESS_INFO, "UTILITYNAME={};VERSION={};".format(
            *get_version_info()))

    def _initiate(self):
        log.info("Export", "Initiating Teradata PT request (awaiting server)  ...")
        start_time = time.time()
        self.export.initiate()
        self.idle_time = time.time() - start_time
        self.initiated = True
        log.info("Export", "Teradata PT request accepted.")
        log.info("Export", "Executing ...")
        log.info(self.options)

    def _fetchall(self, encoding, coerce_floats=None, processor=None):
        if self.query is None:
            raise GiraffeError("Must set target table or query.")
        if not self.initiated:
            self._initiate()
        self.export.set_encoding(encoding)
        if processor is None:
            processor =  identity
        if coerce_floats is None:
            coerce_floats = self.coerce_floats
        if coerce_floats:
            self.export.set_encoding(DECIMAL_AS_FLOAT)
        while True:
            data = self.export.get_buffer()
            if not data:
                break
            for row in data:
                yield processor(row)


class BulkExport(Context):
    __instance__ = TeradataBulkExport



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from collections import defaultdict

from .constants import *
from .errors import *

from ._teradata import Cmd as _Cmd, RequestEnded, StatementEnded, StatementInfoEnded, TeradataError

from .connection import Connection, Context
from .encoders import check_input, null_handler, python_to_sql
from .fmt import format_indent, truncate
from .io import CSVReader, JSONReader, Reader, isfile
from .logging import log
from .sql import parse_statement, prepare_statement, Statement
from .types import Columns, Row
from .utils import pipeline, suppress_context

from ._compat import *


__all__ = ['Cmd', 'Cursor']

_columns_cache = {}
_columns_cache_lock = threading.Lock()


class Cursor(object):
    """
    The class returned by :meth:`giraffez.Cmd.execute <giraffez.cmd.Cmd.execute>` for iterating
    through Terdata CLIv2 results.

    :param `giraffez.Cmd` conn: The underlying database connection
    :param str command: The SQL command to be executed
    :param bool multi_statement: Execute in parallel statement mode
    :param bool header: If :code:`True`, yields a list of column names between statements
    :param bool prepare_only: Execute in prepare mode
    :param bool coerce_floats: Coerce Teradata Decimal types
        automatically into Python floats
    :param bool parse_dates: Returns date/time types as giraffez
        date/time types (instead of Python strings)
    """

    def __init__(self, conn, command, multi_statement=False, header=False,
            prepare_only=False, coerce_floats=True, parse_dates=False,
            panic=True):
        self.conn = conn
        self.command = command
        self.multi_statement = multi_statement
        self.header = header
        self.prepare_only = prepare_only
        self.coerce_floats = coerce_floats
        self.parse_dates = parse_dates
        self.panic = panic
        self.processor = lambda x, y: Row(x, y)
        if not self.coerce_floats:
            self.conn.set_encoding(DECIMAL_AS_STRING)
        if self.parse_dates:
            self.conn.set_encoding(DATETIME_AS_GIRAFFE_TYPES)
        self.columns = None
        if self.multi_statement:
            self.statements = [Statement(command)]
        else:
            self.statements = parse_statement(command)
        self._cur = 0
        self._execute(self.statements[self._cur])
        if self.prepare_only:
            self.columns = self._columns()

    def _columns(self):
        columns = self.conn.columns()
        log.debug("Debug[2]", self.conn.columns(debug=True))
        for column in columns:
            log.verbose("Debug[1]", repr(column))
        return columns

    def _execute(self, statement):
        with statement:
            try:
                self.conn.execute(statement, prepare_only=self.prepare_only)
            except TeradataError as error:
                if self.panic:
                    message = "Statement:\n{}".format(format_indent(truncate(statement), indent="  ", initial="  "))
                    raise suppress_context(TeradataError("{}\n\n{} ".format(error, message)))
                statement._error = error

    def _fetchone(self):
        try:
            row = self.conn.fetchone()
            self.statements[self._cur].count += 1
            return row
        except TeradataError as error:
            if error.code == TD_ERROR_REQUEST_EXHAUSTED:
                if self.multi_statement:
                    return self._fetchone()
                # In some cases CLIv2 returns RequestExhausted instead of
                # StatementEnded. For example, when the statement caused
                # a Teradata syntax error, instead of receiving a
                # StatementEnded parcel, it will return this error when
                # attempting to fetch the next parcel.
                if self._cur == len(self.statements)-1:
                    raise StopIteration
                self._cur += 1
                self._execute(self.statements[self._cur])
                return self._fetchone()
            raise
        except StatementInfoEnded:
            # Indicates that new columns were received so get them and
            # fetch the next parcel.
            self.columns = self._columns()
            self.statements[self._cur].columns = self.columns
            if self.header:
                return self.processor(self.columns, self.columns.names)
            return self._fetchone()
        except StatementEnded:
            if self.multi_statement:
                return self._fetchone()
            # Statement has no more parcels so we fetch the next one
            # to determine if there might be more statements or if the
            # request might be closed.
            if self._cur == len(self.statements)-1:
                raise StopIteration
            self._cur += 1
            self._execute(self.statements[self._cur])
            return self._fetchone()
        except RequestEnded:
            raise StopIteration

    def readall(self):
        """
        Exhausts the current connection by iterating over all rows and
        returning the total.

        .. code-block:: python

            with giraffez.Cmd() as cmd:
                results = cmd.execute("select * from dbc.dbcinfo")
                print(results.readall())
        """
        n = 0
        for row in self:
            n += 1
        return n

    def to_dict(self):
        """
        Sets the current encoder output to Python `dict` and returns
        the cursor.  This makes it possible to set the output encoding
        and iterate over the results:

        .. code-block:: python

            with giraffez.Cmd() as cmd:
                for row in cmd.execute(query).to_dict():
                    print(row)

        Or can be passed as a parameter to an object that consumes an iterator:

        .. code-block:: python

            result = cmd.execute(query)
            list(result.to_dict())
        """
        self.conn.set_encoding(ROW_ENCODING_DICT)
        self.processor = lambda x, y: y
        return self

    def to_list(self):
        """
        Set the current encoder output to :class:`giraffez.Row` objects
        and returns the cursor.  This is the default value so it is not
        necessary to select this unless the encoder settings have been
        changed already.
        """
        self.conn.set_encoding(ROW_ENCODING_LIST)
        self.processor = lambda x, y: Row(x, y)
        return self

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self

    def __next__(self):
        if self.prepare_only:
            raise StopIteration
        data = self._fetchone()
        return self.processor(self.columns, data)

    def __repr__(self):
        return "Cursor(statements={}, multi_statement={}, prepare={}, coerce_floats={}, parse_dates={}, panic={})".format(
            len(self.statements), self.multi_statement, self.prepare_only, self.coerce_floats,
            self.parse_dates, self.panic)


class TeradataCmd(Connection):
    """
    The class for connecting to Teradata and executing commands and 
    queries using CLIv2. 

    Exposed under the alias :class:`giraffez.Cmd`.
    
    For large-output queries, :class:`giraffez.BulkExport <giraffez.export.TeradataBulkExport>` should be used.

    :param str host: Omit to read from :code:`~/.girafferc` configuration file.
    :param str username: Omit to read from :code:`~/.girafferc` configuration file.
    :param str password: Omit to read from :code:`~/.girafferc` configuration file.
    :param int log_level: Specify the desired level of output from the job.
        Possible values are :code:`giraffez.SILENCE` :code:`giraffez.INFO` (default),
        :code:`giraffez.VERBOSE` and :code:`giraffez.DEBUG`
    :param str config: Specify an alternate configuration file to be read from, when
        previous paramaters are omitted.
    :param str key_file: Specify an alternate key file to use for configuration decryption
    :param string dsn: Specify a connection name from the configuration file to be
        used, in place of the default.
    :param bool protect: If authentication with Teradata fails and :code:`protect` is :code:`True` 
        locks the connection used in the configuration file. This can be unlocked using the
        command :code:`giraffez config --unlock <connection>` changing the connection password,
        or via the :meth:`~giraffez.config.Config.unlock_connection` method.
    :param string silent: Suppress log output. Used internally only.
    :param bool panic: If :code:`True`, when an error is encountered it will be
        raised.
    :raises `giraffez.errors.InvalidCredentialsError`: if the supplied credentials are incorrect
    :raises `giraffez.TeradataError`: if the connection cannot be established

    Meant to be used, where possible, with python's :code:`with` context handler
    to guarantee that connections will be closed gracefully when operation
    is complete:

    .. code-block:: python

       with giraffez.Cmd() as cmd:
           results = cmd.execute('select * from dbc.dbcinfo')
           # continue executing statements and processing results

    Using the ``with`` context ensures proper exit-handling and disconnection.
    """

    def __init__(self, host=None, username=None, password=None, log_level=INFO, config=None,
            key_file=None, dsn=None, protect=False, silent=False, panic=True):
        super(TeradataCmd, self).__init__(host, username, password, log_level, config, key_file,
            dsn, protect, silent=silent)
        self.panic = panic
        self.silent = silent

    def execute(self, command, coerce_floats=True, parse_dates=False, header=False, sanitize=True,
            silent=False, panic=None,  multi_statement=False, prepare_only=False):
        """
        Execute commands using CLIv2.

        :param str command: The SQL command to be executed
        :param bool coerce_floats: Coerce Teradata decimal types into Python floats
        :param bool parse_dates: Parses Teradata datetime types into Python datetimes
        :param bool header: Include row header
        :param bool sanitize: Whether or not to call :func:`~giraffez.sql.prepare_statement`
            on the command
        :param bool silent: Silence console logging (within this function only)
        :param bool panic: If :code:`True`, when an error is encountered it will be
            raised.
        :param bool multi_statement: Execute in multi-statement mode
        :param bool prepare_only: Only prepare the command (no results)
        :return: a cursor over the results of each statement in the command
        :rtype: :class:`~giraffez.cmd.Cursor`
        :raises `giraffez.TeradataError`: if the query is invalid
        :raises `giraffez.errors.GiraffeError`: if the return data could not be decoded
        """
        if panic is None:
            panic = self.panic
        self.options("panic", panic)
        self.options("multi-statement mode", multi_statement, 3)
        if isfile(command):
            self.options("file", command, 2)
            with open(command, 'r') as f:
                command = f.read()
        else:
            if log.level >= VERBOSE:
                self.options("query", command, 2)
            else:
                self.options("query", truncate(command), 2)
        if not silent and not self.silent:
            log.info("Command", "Executing ...")
            log.info(self.options)
        if sanitize:
            command = prepare_statement(command) # accounts for comments and newlines
            log.debug("Debug[2]", "Command (sanitized): {!r}".format(command))
        self.cmd.set_encoding(ENCODER_SETTINGS_DEFAULT)
        return Cursor(self.cmd, command, multi_statement=multi_statement, header=header,
            prepare_only=prepare_only, coerce_floats=coerce_floats, parse_dates=parse_dates,
            panic=panic)

    def exists(self, object_name, silent=False):
        """
        Check that object (table or view) :code:`object_name` exists, by executing a :code:`show table object_name` query, 
        followed by a :code:`show view object_name` query if :code:`object_name` is not a table.

        :param str object_name: The name of the object to check for existence.
        :param bool silent: Silence console logging (within this function only)
        :return: :code:`True` if the object exists, :code:`False` otherwise.
        :rtype: bool
        """
        try:
            self.execute("show table {}".format(object_name), silent=silent)
            return True
        except TeradataError as error:
            if error.code != TD_ERROR_OBJECT_NOT_TABLE:
                return False
        try:
            self.execute("show view {}".format(object_name), silent=silent)
            return True
        except TeradataError as error:
            if error.code not in [TD_ERROR_OBJECT_NOT_VIEW, TD_ERROR_OBJECT_NOT_EXIST]:
                    return True
        return False

    def fetch_columns(self, table_name, silent=False):
        """
        Return the column information for :code:`table_name` by executing a :code:`select top 1 * from table_name` query.

        :param str table_name: The fully-qualified name of the table to retrieve schema for
        :param bool silent: Silence console logging (within this function only)
        :return: the columns of the table
        :rtype: :class:`~giraffez.types.Columns`
        """
        return self.execute("select top 1 * from {}".format(table_name), silent=silent, prepare_only=True).columns

    def insert(self, table_name, rows, fields=None, delimiter=None, null='NULL', parse_dates=False, quotechar='"'):
        """
        Load a text file into the specified :code:`table_name` or Insert Python :code:`list` rows into the specified :code:`table_name`

        :param str table_name: The name of the destination table
        :param list/str rows: A list of rows **or** the name of an input file. Each row
            must be a :code:`list` of field values.
        :param list fields: The names of the target fields, in the order that
            the data will be presented (defaults to :code:`None` for all columns
            in the table).
        :param str delimiter: The delimiter used by the input file (or :code:`None`
            to infer it from the header).
        :param str null: The string used to indicated nulled values in the
            file (defaults to :code:`'NULL'`).
        :param str quotechar: The character used to quote fields containing special
            characters, like the delimiter.
        :param bool parse_dates: If :code:`True`, attempts to coerce date fields
            into a standard format (defaults to :code:`False`).
        :raises `giraffez.errors.GiraffeEncodeError`: if the number of values in a row does not match
            the length of :code:`fields`
        :raises `giraffez.errors.GiraffeError`: if :code:`panic` is set and the insert statement
            caused an error.
        :return: A dictionary containing counts of applied rows and errors
        :rtype: :class:`dict`

        For most insertions, this will be faster and produce less strain on
        Teradata than using :class:`~giraffez.load.TeradataBulkLoad` (:class:`giraffez.BulkLoad <giraffez.load.TeradataBulkLoad>`).

        Requires that any input file be a properly delimited text file, with a
        header that corresponds to the target fields for insertion. Valid delimiters
        include '|', ',', and <tab> or a properly encoded JSON stream.
        """
        if not isfile(rows):
            return self._insert(table_name, rows, fields, parse_dates)
        with Reader(rows, delimiter=delimiter, quotechar=quotechar) as f:
            preprocessor = null_handler(null)
            rows = (preprocessor(l) for l in f)
            if isinstance(f, CSVReader):
                self.options("delimiter", unescape_string(f.reader.dialect.delimiter), 1)
                self.options("quote char", f.reader.dialect.quotechar, 2)
            elif isinstance(f, JSONReader):
                self.options("encoding", "json", 1)
            return self._insert(table_name, rows, f.header, parse_dates)

    def _close(self, exc=None):
        if getattr(self, 'cmd', None):
            self.cmd.close()

    def _connect(self, host, username, password):
        self.cmd = _Cmd(host, username, password)

    def _insert(self, table_name, rows, fields=None, parse_dates=False):
        global _columns_cache
        with _columns_cache_lock:
            if table_name in _columns_cache:
                columns = _columns_cache[table_name]
            else:
                columns = self.fetch_columns(table_name, silent=True)
                if not columns:
                    raise GiraffeError("Fetch columns failed '{}'".format(table_name))
                _columns_cache[table_name] = columns
        if fields is None:
            fields = columns.safe_names
        columns.set_filter(fields)
        check_input(columns, fields)
        stats = defaultdict(int)
        processor = pipeline([
            python_to_sql(table_name, columns, parse_dates)
        ])
        def _fetch():
            stats['count'] = 0
            current_block = ""
            for row in rows:
                try:
                    stmt = processor(row)
                except GiraffeError as error:
                    if self.panic:
                        raise error
                    log.info("Command", error)
                    stats['errors'] += 1
                    continue
                if len(current_block + stmt) > CLI_BLOCK_SIZE:
                    yield current_block
                    current_block = ""
                current_block += stmt
                stats['count'] += 1
            if current_block:
                yield current_block
        log.info("Command", "Executing ...")
        for i, block in enumerate(_fetch()):
            self.execute(block, sanitize=True, multi_statement=True, silent=True)
            if i == 0:
                log.info(self.options)
        return stats

class Cmd(Context):
    __instance__ = TeradataCmd
